;
; miscutil.asm - misc util
;
; Written by
;  Kajtar Zsolt <soci@c64.rulez.org>
;
;    This file is part of IDEDOS the IDE64 disk operating system
;    See README for copyright notice.
;
;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;
;----------------------------------
; Read buffer
;
XIT_read_in_buffer .proc
		php
		jsr XIT_rw_com		;-=> error
                plp
                jsr next_sector0
		beq clear_buffer2	;hole in file?
		jsr OP_read_cmd1
                bcs XIT_read_a_sector.ujj;-=> error

		jsr OP_buffer_start
		tax
		jsr read_buffer_cfs
                #jmps XIT_read_a_sector.ji;-=> error
;----------------------------------
; Clear buffer
;
clear_buffer2   .proc
		bcc clear512.no
clear_buffer    .proc
		jsr OP_buffer_start
clear512	.proc
		jsr init_pom1_y
ns		tya
		jsr nw
nw		#sta_zp_y pom1
		iny
		bne nw
		inc pom2
no		rts
;-----------------------------------
; Read in a sector
;
XIT_read_a_sector .proc
		pha
		jsr OP_read_cmd1
ujj		jsr LWOP_reporterror_read_xit;-=> error
		pla
		jsr OP_read512_here
		jsr OP_get_current_drive_type
		and #$c0
		eor #$80
		bne ji			;cdrom?
		jsr OP_read1536
ji		#jmps XIT_nomoreread;-=> error
		.pend
;-----------------------------------
; Write out a sector
;
XIT_write_a_sector .proc
		pha
		jsr OP_write_cmd1
		bcs XIT_write_out_buffer.vw    ;this branch must be here!
		pla
		bne nonempty
		tay
		sty low_reg
		sty hig_reg
ni		sta data_g
		sta data_g
		iny
		bpl ni
		gmi XIT_write_out_buffer.in

nonempty	jsr init_pom1_y

		sec
		.byte SKIP1
lp		clc
lp2		#lda_zp_y pom1
		sta low_reg		;0-7 BIT
		iny

		#lda_zp_y pom1
		sta hig_reg		;8-15 BIT
		sta data_g
		iny
		bne lp2
		inc pom2
		bcs lp
		gcc XIT_write_out_buffer.in
;----------------------------------
; Write buffer
;
XIT_write_out_buffer .proc
		jsr XIT_rw_com		;-=> error
		jsr OP_write_cmd1
                bcs vw                  ;-=> error

		jsr OP_buffer_start
		tax
		jsr write_buffer_cfs
in		jsr ata_pi_nomore_data
		bcs vw
		jsr wait_busy
vw		jmp LWOP_reporterror_write_xit;-=> error
		.pend
;----------------------------------
; Buffer common
;
XIT_rw_com	.proc
		jsr buffer_tree_pom1
		ldy #3
ce		#lda_zp_y pom1
		sta next_sector,y
                dey
                bpl ce
		rts
		.pend
;-----------------------------------
; Loads a tree sector and create if non-existent
;
XIT_load_tree	.proc
		#lda0
		.byte SKIP2
;-----------------------------------
; Loads a tree sector (walks tree)
;
XIT_save_tree	.proc
m		= pom4
m2		= pom6
level		= pom8

		lda #1
		pha

		#ldx_l_a buffer_currentp_exp
		lda buffer_filepos4_exp,x
                sta m+1
                cmp buffer_treeposhi_exp,x
                sta buffer_treeposhi_exp,x
                php
                lda buffer_filepos3_exp,x;need to search for a new tree sector?
		sta m
		cmp buffer_treeposlo_exp,x
                sta buffer_treeposlo_exp,x;save current fp
		bne do
		plp
		bne do2
		lda buffer_treestate_exp,x
                beq do2			;empty, force!
		pla
		rts			;-=> ok

do		plp
do2		jsr restore_file_state

		jsr XIT_flush_tree

		#cnext cfs_file_treeroot_exp , next_sector

		ldy #0
		sty m2
		sty m2+1

		lda m
idfy		tax
		ora m+1
		beq en

		txa
		bne *+4
		dec m+1
		dex
		txa

		lsr m+1
		ror
		rol m2
		rol m2+1
		lsr m+1
		ror
		rol m2
		rol m2+1
		lsr m+1
		ror
		rol m2
		rol m2+1

		iny
		gne idfy

ki		pla
		jmp save_file_state

en		sty level

loop		jsr OP_buffer_start
		ora #>b_cfs_file_tree_exp
		jsr XIT_read_a_sector	;-=> error

		#ldx_l_a buffer_currentp_exp
                lda #$01
		sta buffer_treestate_exp,x;mark tree sector ok

loop2		#cnext next_sector , cfs_file_tree_exp;remember tree sector position

		dec level
		bmi ki

		#lda0
		tay
		lsr m2+1
		ror m2
		ror
		lsr m2+1
		ror m2
		ror
		lsr m2+1
		ror m2
		sta pom1
		jsr OP_buffer_start
		adc #>b_cfs_file_tree_exp
		sta pom2
		ldx #3
mic		#lda_zp_y pom1
		asl
		asl
		asl
		rol next_sector,x
		asl
		rol next_sector,x
		tya
		clc
		adc #4
		tay
		and #15
		bne mic
		dex
		bpl mic
		jsr next_sector0
		bne loop

		pla		;mode?
		beq hole
                pha
                
                #psz pom1

                lda #FIND_FREE_METADATA
		jsr find_free_space

		txa
		jsr XIT_write_a_sector	;save empty tree sector

		pla
                sta pom1
                pla
                sta pom2

                #ldx_l_a buffer_currentp_exp
                lda #$80
		sta buffer_treestate_exp,x;mark tree sector dirty

		ldy #0
		ldx #3
mi2             lda next_sector,x
                pha			;remember
mi		jsr updatelink
		adc #4
		tay
		and #15
		bne mi
		dex
		bpl mi2

		jsr XIT_flush_free_and_tree	;flush old tree sector

		jsr clrtree		;empty tree sector

		ldx #0
cqe             pla
		sta next_sector,x	;restore
		inx
                cpx #4
                bcc cqe
		gcs loop2

hole		jsr save_file_state
		#ldx_l_a buffer_currentp_exp
		#lda0			;not sector not available
		sta buffer_treestate_exp,x

clrtree 	jsr buffertreepom1
		#jmps clear512            ;clear tree sector

buffertreepom1 	jsr OP_buffer_start
		ora #>b_cfs_file_tree_exp
                jmp init_pom1_y
                .pend

XIT_flush_free_and_tree	.proc
		jsr XIT_find_free_flush	;flush bam

XIT_flush_tree	.proc
		#ldx_l_a buffer_currentp_exp
                lda buffer_treestate_exp,x
                bpl eu		        ;OK, not dirty
		
                lda #1
		sta buffer_treestate_exp,x

		#cnext cfs_file_tree_exp , next_sector ;restore tree sector
                
                jsr OP_buffer_start
		ora #>b_cfs_file_tree_exp;write out the tree sector
		jmp XIT_write_a_sector	;-=> error

eu		rts
		.pend

updatelink      #lda_zp_y pom1
		asl
                rol
                rol
		asl next_sector,x
		rol
		asl next_sector,x	;update link
		rol
		rol
		rol
		rol
		rol
		#sta_zp_y pom1
                tya
		rts
;--------------------------------------
;Init drive (eliminate possible diskchange problems...)
;--------------------------------------
;Before drive access
;
XIT_init_drive2	.proc
		clv
in		jsr OP_get_current_drive_type
		bpl noatapi
                asl
                bpl *+3			;CDROM only
                clv
                php
                
		lda #ATAPI_TEST_UNIT_READY
		jsr OP_atapi_packet
		jsr OP_atapi_command
                rol
                plp
		cmp #EATAPIERROR*2+1
		bne n
		lda error_g
		#lda_low_reg
                and #$f0
		cmp #ATAPI_SENSE_UNIT_ATTENTION
		bne reqsense
inv		jsr invalidate_drive
n		jsr OP_get_current_drive_type
noatapi		rts

reqsense	bvc n
		lda #ATAPI_MODE_SENSE
		jsr OP_atapi_packet
                #pc 2,1
                #pc 8,4
                jsr OP_atapi_datain_command
                bcs n
                lda data_g
                lda data_g
                #lda_low_reg
                cmp #$70		;Buggy CDROM...
                beq inv
                cmp #$71
                bne n
                jsr invalidate_drive	;Buggy CDROM...
                lda #ATAPI_CMD_START_STOP
                jsr OP_atapi_packet
                #pc 4,3
                jsr OP_atapi_command
	        #jmps XIT_init_drive2
		.pend
;--------------------------------
;Before filesystem access
;
XIT_init_drive	.proc
		#sev v
		jsr XIT_init_drive2.in
		and #FS_MASK
		clc
		beq OP_XIT_init_filesystem_from_disk
v       	rts
		.pend
;--------------------------------------
;Calculate linear access from tree sector
;--------------------------------------
linearaccess	.proc
		.if SUPERCPU=1
		lda pom3
		pha
		lda #1
		sta pom3
                .fi

		#psz pom1	;store pom2, pom1

		lda load_p_exp
                pha
		asl
		asl
		sta pom1
                tya
                adc #0		;now C=0
		sta pom2

		ldy #3
vcs		#lda_zp_y pom1
		sta next_sector,y
		dey
		bpl vcs

		#ldx_l_a current_drive_exp
                tya
		adc load_size	;how many sectors left?
		lda #>511
		adc load_size+1
		ror
		cmp setup_linear_r_hdd0_exp,x
		bcc idi
		beq idi
		#lda0		;lda #1-1
		adc setup_linear_r_hdd0_exp,x
idi		sta load_limit_exp

linlp		#inc_l_a load_p_exp
		bmi gy

		#dec_l_a load_limit_exp
		beq gy

		ldy #3
		bit next_sector
		bvc chs

		sec
                #lda0
		#adc_zp_y pom1
		ldy #7
		#eor_zp_y pom1
		bne gy
		ldy #2
		#adc_zp_y pom1
		ldy #6
		#eor_zp_y pom1
		bne gy
		ldy #1
		#adc_zp_y pom1
		ldy #5
		#eor_zp_y pom1
		bne gy
		tay
		#adc_zp_y pom1
		ldy #4
		#eor_zp_y pom1
		and #15
		beq vrg

gy		pla
		clc
                sbc load_p_exp
                eor #255
		sta load_sectors_exp

		pla
		sta pom1
		pla
		sta pom2
		.if SUPERCPU=1
		pla
		sta pom3
		.fi
		rts

chs		#lda_zp_y pom1		;sector
		cmp drive_maxsector_exp,x
		bcc *+4
		lda #255
		adc #1
		ldy #7
		#eor_zp_y pom1
		bne gy
		tay			;head
		#lda_zp_y pom1
		bcc vk1
		and #15
		cmp drive_maxhead_exp,x
		bcc *+4
		lda #254
		adc #1
vk1		ldy #4
		#eor_zp_y pom1
		and #15
		bne gy
		ldy #2			;cyllow
		#adc_zp_y pom1
		ldy #6
		#eor_zp_y pom1
		bne gy
		ldy #1			;cylhi
		#adc_zp_y pom1
		ldy #5
		#eor_zp_y pom1
		bne gy

vrg		lda pom1
		clc
		adc #4
		sta pom1
		bcc linlpx
		inc pom2
linlpx		jmp linlp
		.pend
;----------------------------------
; Calculate the bam sector position for the sector in next_sector
; result in next_sector
;
find_free_start	.proc
szam_thiss	= szam+5

		sta find_free_buf_exp+1	;this buffer is used for bam
		sta find_free_bufmeta_exp+1
		#lda0
		sta find_free_dirty_exp	;it's not dirty

		jsr chs2lba

		ldx #3
ll		lda szam,x
		sta szam_thiss,x
		dex
		bpl ll

		#ldx_l_a partition_currentp_exp
		lda partition_root_dir_exp,x
		asl
		php			;save which is the active bitmap
		#lda0
                sta find_free_direction_exp;forward search
		rol
		sta find_free_bam_exp

                jsr partitionboundprepare

		lda szam_thiss+0
		cmp szam+0
		ldx #0
		stx szam_thiss+0
		lda szam_thiss+1
		sbc szam+1
		and #$f0
		sta szam_thiss+1
		lda szam_thiss+2
		sbc szam+2
		sta szam_thiss+2
		lda szam_thiss+3
		sbc szam+3
		sta szam_thiss+3

		ldy #4
		plp
li		lda szam,x
		adc szam_thiss,x
		sta szam,x
		inx
		dey
		bne li
                
                jsr findnextbam.validate
;----------------------------------
; Read in bam sector
;
readinbam2     	.proc
		lda find_free_direction_exp
                asl
                #lda0
                sta find_free_bufmeta_exp
                bcs baj
		lda #METADATA_RESERVED
baj             sta find_free_buf_exp

readinbam      	.proc

		#cnext szam , find_free_now_exp

in		jsr lba2chs

		lda find_free_buf_exp+1
		and #%11111110
		sta find_free_buf_exp+1
		sta find_free_bufmeta_exp+1
		pha
		jsr XIT_read_a_sector	;-=> error
                pla
                jsr init_pom1_y
                #lda_zp_y pom1
                and #%00111111		;make sure BAM is always allocated!
                #sta_zp_y pom1
                rts			;Y=0! N=0!
		.pend
;----------------------------------
; Calculate partitionstart to lba number
;
partitionboundprepare .proc
		clc
                jsr tk
                sec
tk		#ldx_l_a partition_currentp_exp
		ldy #0
lq2		lda partition_start_exp,x
		bcs ti
		lda partition_end_exp,x
ti		sta next_sector,y
		inx
                iny
                tya
                and #3
                bne lq2
                php
		jsr chs2lba
                plp

                ldx #3
xei             lda szam,x
                sta find_free_start_exp,x
                bcs e2
                sta find_free_end_exp,x
e2		dex
		bpl xei
                rts
		.pend
;-----------------------------------
; Find and allocate a sector(s)
;
find_free_space .proc
szam_offs	= szam+5
		sta find_free_mode_exp
		lda #1
		sta load_limit_exp
lotof
		lda find_free_direction_exp
                and #1
                sta find_free_direction_exp

                #ldx_l_a find_free_mode_exp
		lda find_free_buf_exp+1,x
                jsr init_pom1_y

retr		#ldy_l_a find_free_buf_exp,x
		#jmps id

id2		rol
		sta find_free_buf_exp+1,x
id		#lda_zp_y pom1
		bne foundfree
		iny
		bne id
		inc pom2
		lda pom2
		lsr
		bcs id2

		jsr XIT_find_free_flush	;-=> error

		jsr find_free_now_exp_to_szam

                jsr findnextbam
                jsr readinbam2
                gpl retr

foundfree      	sta find_free_dirty_exp
		ldx #255
xxu		inx
		asl
		bcc xxu
		pha
		stx szam_offs
		lda pom2
		and #1
		sta szam_offs+1
		tya
		asl
		rol szam_offs+1
		asl
		rol szam_offs+1
		asl			;x8
		rol szam_offs+1
		ora szam_offs
		sec
		sbc find_free_bam_exp	;correction if bam!=0
		sta szam_offs
		bcs cvs
		dec szam_offs+1
cvs
		pla

allp		#inc_l write_p_exp
		bmi endlin

		#dec_l load_limit_exp
		beq endlin

		inx
ce		asl
		bcs allp
		cpx #8
		bcc endlin

		#sta_zp_y pom1

		iny
		bne nbord
                #ldx_l_a find_free_mode_exp
		inc pom2
		lda pom2
		lsr
                dey
		bcc endofbam
                iny
		rol
		sta find_free_buf_exp+1,x

nbord		#lda_zp_y pom1
		ldx #0
		geq ce

endlin		pha
		#sbxc_a 8
		pla
                			;C is 0
vm		rol
		inx
		bne vm

		#sta_zp_y pom1
                
                #ldx_l_a find_free_mode_exp

endofbam       	tya
		sta find_free_buf_exp,x

		jsr find_free_now_exp_to_szam

		lda szam
		clc
		adc szam_offs
		sta szam
		lda szam+1
		adc szam_offs+1
		sta szam+1
		bcc vd
		inc szam+2
		bne vd
		inc szam+3
qqqq            sec
                beq LWOP_reporterror_72_xit;disk full
vd		jsr boundend
		bcs qqqq		;disk full
                jsr boundstart
                bcc qqqq		;disk full
		jmp lba2chs
		.pend
;-----------------------------------
; Validate that we are in valid sector range
;
boundend       	.proc
		ldx #0
                ldy #4
                sec
ku		lda szam,x
                sbc find_free_end_exp,x
                inx
                dey
                bne ku
                rts
                .pend
;---------
boundstart      .proc
		ldx #0
                ldy #4
                sec
ku2		lda szam,x
                sbc find_free_start_exp,x
                inx
                dey
                bne ku2
                rts
		.pend
;-----------------------------------
; Find next bam sector
;
findnextbam     .proc
		lda find_free_direction_exp
lpnkq           lsr
                bcc dir
	        lda szam+1		;next block
		sec
		sbc #16
		sta szam+1
		bcs uj
		lda szam+2
                bne nj
		lda szam+3
                beq df
		dec szam+3
nj		dec szam+2
uj
		jsr boundend
                bcs df
		jsr boundstart
                bcs validate
                ldx #$80		;opposite direction
                gne ntkq
                
df              sec
		gcs LWOP_reporterror_72_xit;disk full
		
dir		lda szam+1		;next block
		clc
		adc #16
		sta szam+1
		bcc ue2
		inc szam+2
		bne ue2
		inc szam+3
	       	beq df			;can't happen

ue2		jsr boundstart
                bcc df
		jsr boundend
                bcc validate
                ldx #$81		;opposite direction
ntkq            lda find_free_direction_exp
		bmi df			;second direction change!
                txa
		sta find_free_direction_exp
                gne lpnkq
validate
		jsr isbamlocked
                bcs findnextbam		;yes...
                rts
                .pend

isbamlocked	.proc
                .if SUPERCPU=1
                lda #1
                sta next_sector+2
                .fi
                lda #<(b_cfs_file_state_exp+find_free_now_exp-cfs_file_info_backup_start_exp)
                sta next_sector

		ldx #MAX_BUFFERS-1
csc		lda buffer_logical_exp,x
                beq nxt		       ;buffer valid?
                lda buffer_drive_exp,x
                cmp current_drive_exp
                bne nxt  		;on this drive?
                lda buffer_type_exp,x
                cmp #BUFFERMODE_FILE
                bne nxt
		lda buffer_info_exp,x
                and #INFO_NON_WRITEABLE
                bne nxt
                lda buffer_start,x
                ora #>b_cfs_file_state_exp
                sta next_sector+1
                ldy #3
llu             #lda_zp_y next_sector
                eor szam,y
                bne nxt
                dey
                bpl llu
                sec
                rts
nxt		dex 
		bpl csc
                clc
		rts
		.pend
;-----------------------------------
; Flush bam
;
XIT_find_free_flush .proc
		lda find_free_dirty_exp
		beq non			;not dirty, no flush

		jsr find_free_now_exp_to_szam
                
                jsr lba2chs

		lda find_free_buf_exp+1
		and #%11111110
		jsr XIT_write_a_sector

		#lda0
		sta find_free_dirty_exp
non		rts
		.pend
;-----------------------------------
; Find free sectors and fill in datatree
; A: tree buffer
; Y: 0 truncate, 1 truncate but one more sector, 2 round up
;
find_free_sectors_blockwrite .proc
p1		= current_path
p2		= current_path+1
p3		= current_path+2

		#ldx_l_a current_drive_exp
		lda write_size+1
                clc
		adc write_p1_exp
		ror
		cmp setup_linear_w_hdd0_exp,x
		bcc idi
		beq idi
		#lda0		;lda #1-1
		adc setup_linear_w_hdd0_exp,x
idi		sta load_limit_exp

		lda write_p_exp
		pha
                jsr restore_file_state
		lda #FIND_FREE_DATA
                sta find_free_mode_exp
		jsr find_free_space.lotof
                jsr save_file_state
		pla
                sta p1
		clc
		sbc write_p_exp
		eor #255
		sta write_sectors_exp
		sta load_limit_exp

		#ldx_l_a buffer_currentp_exp
		lda #$80		;dirty
                sta buffer_treestate_exp,x
                
		asl p1
		asl p1
		lda buffer_start,x
		adc #>b_cfs_file_tree_exp;now C=0
		sta p2
		.if SUPERCPU=1
		lda #1
		sta p3
		.fi

		ldy #3
ncs		lda next_sector,y
		#sta_zp_y p1
		dey
		bpl ncs
                
                #ldx_l_a current_drive_exp

lp		#dec_l_a load_limit_exp
		beq ex

		ldy #3
		#lda_zp_y p1
		bit next_sector
		bvc chs

		clc
		adc #1
		ldy #7
		#sta_zp_y p1
		ldy #2
                #lda0
		#adc_zp_y p1
		ldy #6
		#sta_zp_y p1
		ldy #1
		#lda0
		#adc_zp_y p1
		ldy #5
		#sta_zp_y p1
		#lda0
		tay
		#adc_zp_y p1
		ldy #4
		gne cxm

ex      	rts

chs		cmp drive_maxsector_exp,x
		bcc *+4
		lda #255
		adc #1
		ldy #7
		#sta_zp_y p1
		ldy #0
		#lda_zp_y p1
		and #15
		bcc md
		cmp drive_maxhead_exp,x
		bcc *+4
		lda #254
		adc #1
md		ldy #4
		#sta_zp_y p1
		ldy #2
		#lda0
		#adc_zp_y p1
		ldy #6
		#sta_zp_y p1
		ldy #1
		#lda0
		#adc_zp_y p1
		ldy #5
cxm		#sta_zp_y p1

		lda p1
		clc
		adc #4
		sta p1
		bcc lp
		inc p2
		gcs lp
		.pend
;-----------------------------------
; Convert next_sector -> szam
;
chs2lba		.proc
		bit next_sector
		bvc konv
		lda next_sector+0
		and #%01001111
		sta szam+3
		lda next_sector+1
		sta szam+2
		lda next_sector+2
		sta szam+1
		lda next_sector+3
		sta szam+0
		rts

konv		lda next_sector+2
		sta szam
		lda next_sector+1
		sta szam+1
		#stz_a szam+2
		#stz_aa szam+3

		#ldx_l_a current_drive_exp
		lda drive_maxsector_exp,x
		pha
		lda drive_maxhead_exp,x
		tax
		inx
		jsr mul			;cyl*maxhead

		lda next_sector+0
		and #15
		jsr sum			;cyl*maxhead+head

		#plx_a
		jsr mul			;(cyl*maxhead+head)*maxsector
		jsr mul_p4bit

		lda next_sector+3
                #sbc1c
		jmp sum			;(cyl*maxhead+head)*maxsector+sector-1
		.pend
;-----------------------------------
; Convert szam -> next_sector
;
lba2chs		.proc
		bit szam+3
		bvc konv
		lda szam+3
		sta next_sector+0
		lda szam+2
		sta next_sector+1
		lda szam+1
		sta next_sector+2
		lda szam+0
		sta next_sector+3
		rts

konv		#ldx_l_a current_drive_exp
		lda drive_maxhead_exp,x
		pha
		lda drive_maxsector_exp,x
		tax
		jsr div
                #adc1c
		sta next_sector+3

		#plx_a
		inx
		jsr div
		sta next_sector
		lda szam
		sta next_sector+2
		lda szam+1
		sta next_sector+1
		rts
		.pend
;-----------------------------------
; Read directory sector/allocate new
; A=bitmap buffer
; X=bonus illegal character (eg. "/" in dirname)
;
XIT_prealloc_dir_sector .proc
		pha
                #stx_l_a wildcard_end_exp

		jsr XIT_wp

		ldx #0
nw		lda path_exp,x
		cmp wildcard_end_exp
                beq sw
		jsr is_special_s
		beq sw
		inx
		gne nw
sw		cmp #1			;0 ?
		jsr LWOP_reporterror_33_xit

		ldx #$fd
nw2		lda path_searchtype_exp-$fd,x
		cmp #32			;no space in type!
		beq sw2
                cmp #"<"		;no < in type!
                beq sw2
                cmp #"/"		;no / in type!
                beq sw2
		jsr is_special_s
		beq sw2
		inx
		bne nw2
		txa			;ok, none found
sw2		cmp #1			;0 ?
		bcs sw			;LWOP_reporterror_33_xit

		#cnext resolv_this_sector_exp , next_sector;get last read dir sector

		pla
		jsr find_free_start	;start allocation from last read dir sector

		lda resolv_empty_entry_found_exp
                beq nospace
                rts

nospace 	lda path_dirsectors_exp
                cmp #DIRECTORY_LENGTH
                jsr LWOP_reporterror_72_xit;sorry, no more directory entrys!

                lda #FIND_FREE_METADATA
		jsr find_free_space	;allocate next directory sector

		#cnext next_sector , resolv_empty_entry_exp

		#lda0			;have to write out new empty dir sector first
                sta resolv_entry_offs_exp;it's the first entry
		jsr XIT_write_a_sector	;-=> error

		lda #ram_exp_dir
		jsr init_pom1_y

		lda #$14
                sta pom1
		ldx #3
mve		jsr updatelink
		adc #32
		tay
                and #$7f
                bne mve
                dex
		bcc mve
		inc pom2
                txa
                bpl mve

		jsr XIT_find_free_flush	;-=> error we must first flush the bam, then extend the directory!

		#cnext resolv_this_sector_exp , next_sector

		lda #ram_exp_dir	;write out modified dir sector
		jsr XIT_write_a_sector	;-=> error

		lda #ram_exp_dir	;clear new dir sector in memory too
                #jmps clear512          ;now we have a longer directory with free entrys...
;-----------------------------------
; Fill in directory entry info
;
XIT_finalize_direntry .proc
		pha
                php
                lda resolv_entry_offs_exp
                lsr
                ror
                ror
                ror
                sta pom1
                lda #ram_exp_dir/2
                rol
                sta pom2
                .if SUPERCPU=1
                ldx #1
                stx pom3
                dex
                .else
                ldx #0
                .fi
                
                ldy #$10
                plp
                bcc eu

		#txy_a
ii		lda path_exp,x		;fill in filename
		beq *+3
		inx
		#sta_zp_y pom1
		iny
		cpy #$10
		bcc ii

eu      	ldx #4
mc		lda szam-1,x
		#sta_zp_y pom1
		iny
		dex
		bne mc			;fill in filesize
                bcc eu2

		#lda_zp_y pom1
		and #%00110000		;nexts
fv		#ora_l_y_x cfs_file_treeroot_exp-$14
		#sta_zp_y pom1		;fill in tree sector
                .if SUPERCPU=1
                tdc
                .else
                txa
                .fi
		iny
		cpy #$18
		bcc fv

eu2		pla			;fileattrib
		bcs eu3
                ldy #$18
                #ora_zp_y pom1
eu3		#sta_zp_y pom1
		iny
                bcc eu4

scw		#lda_l_y_x path_searchtype_exp-$19
		#sta_zp_y pom1		;fill in attributes and filetype
		iny
		cpy #$1c
		bcc scw

		#cnext resolv_empty_entry_exp , next_sector

eu4		jsr OP_timestamp	;fill in timestamp

		lda #ram_exp_dir	;write out the final directory entry
		jmp XIT_write_a_sector	;-=> error
		.pend

;-----------------------------------
; Update filelength if position is after fileend
; Y not changed
;
update_filelen	.proc
                sec
		#lda0
		adc buffer_address_low_exp,x
		sta pom1
                #lda0
		adc buffer_address_hi_exp,x
		sec
		sbc buffer_start,x
                and #$7f
		clc
		adc buffer_filepos2_exp,x
		sta pom2
		#lda0
		adc buffer_filepos3_exp,x
		sta pom3
		#lda0
		adc buffer_filepos4_exp,x
		sta pom4
		lda buffer_filesize1_exp,x
		cmp pom1
		lda buffer_filesize2_exp,x
		sbc pom2
		lda buffer_filesize3_exp,x
		sbc pom3
		lda buffer_filesize4_exp,x
		sbc pom4
		bcs gse
		lda pom1
		sta buffer_filesize1_exp,x
		lda pom2
		sta buffer_filesize2_exp,x
		lda pom3
		sta buffer_filesize3_exp,x
		lda pom4
		sta buffer_filesize4_exp,x
gse		rts
		.pend

find_free_now_exp_to_szam .proc
		#cnext find_free_now_exp , szam
                rts
                .pend
